{-# OPTIONS_GHC -F -pgmF htfpp #-}
{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE CPP #-}
{-# LANGUAGE OverloadedStrings #-}
module Mgw.MessageQueue.BrokerServer
    ( createClientHandler, startBrokerServer, brokerServerMain
    , htf_thisModulesTests )
where

#include "src/macros.h"

----------------------------------------
-- LOCAL
----------------------------------------
import Mgw.MessageQueue.LocalBroker
import Mgw.MessageQueue.Types
import Mgw.MessageQueue.Protocol
import Mgw.Util.Concurrent
import Mgw.Util.STM
import Mgw.Util.Tx
import Mgw.Util.Logging hiding (withLogId)
import Mgw.Util.Preview
import Mgw.Util.TestHelper

----------------------------------------
-- SITE-PACKAGES
----------------------------------------
import Data.Conduit
import Data.Conduit.Network
import Safe
import Test.Framework
import qualified Data.Conduit as C
import qualified Data.HashMap.Strict as HashMap
import qualified Data.Text as T
import qualified Data.Vector as V

----------------------------------------
-- STDLIB
----------------------------------------
import Control.Monad.IO.Class
import System.Environment

createClientHandler ::
    LogId
    -> MessageBroker q
    -> IO (C.Source IO ServerMessage, C.Sink ClientMessage IO ())
createClientHandler logId mb =
    do replyChan <- runTx $ newTBMChan ("replyChan_" ++ show logId) 10
       return (source replyChan, sink replyChan)
    where
      logMsg = withLogId logId
      source :: TBMChan ServerMessage -> C.Source IO ServerMessage
      source replyChan =
          do liftIO $ logInfo (logMsg "Sending queue list")
             C.yield (ServerQueues (V.fromList (HashMap.keys (mb_queues mb))))
             let loop =
                     do mSrvMsg <- liftIO $ runTx $ readTBMChan replyChan
                        case mSrvMsg of
                          Just msg -> C.yield msg >> loop
                          Nothing -> return ()
             loop
      sink :: TBMChan ServerMessage -> C.Sink ClientMessage IO ()
      sink replyChan =
          let loop =
                  do mMsg <- C.await
                     case mMsg of
                       Nothing ->
                           liftIO $
                           do logNote (logMsg ("client closed connection"))
                              runTx $ closeTBMChan replyChan
                       Just clMsg ->
                           do liftIO $ handleClientMessage clMsg replyChan
                              loop
          in loop
      handleClientMessage msg replyChan =
          case msg of
            ClientSubscribe queueName ->
                case mb_lookupQueue mb queueName of
                  Nothing ->
                      logWarn (logMsg ("ignoring subscription to unkown queue: " ++
                                       preview queueName))
                  Just q ->
                      do logInfo (logMsg ("received subscription for " ++ preview queueName))
                         mb_subscribeToQueue mb q (messageHandler queueName replyChan)
            ClientPublishMessage queueName msg ->
                case mb_lookupQueue mb queueName of
                  Nothing ->
                      logWarn (logMsg ("ignoring publication of message " ++
                                       preview (msg_id msg) ++ " for unknown queue " ++
                                       preview queueName))
                  Just q ->
                      do logInfo (logMsg ("received new message " ++ preview (msg_id msg) ++
                                          " for queue " ++ preview queueName))
                         mb_publishMessage mb q msg
      messageHandler queueName replyChan =
          mkSubscriber (T.pack $ "handler_" ++ show logId ++ "_" ++ preview queueName) $ \msg ->
                       do logInfo (logMsg ("sending message " ++ preview (msg_id msg) ++
                                           " for queue " ++ preview queueName ++
                                           " to client"))
                          runTx $ writeTBMChan replyChan (ServerPublishMessage queueName msg)


startBrokerServer :: Int -> MessageBroker q -> IO ()
startBrokerServer port mb =
    runTCPServer (serverSettings port "*") $ \appData ->
        do let logId = LogId (T.pack (show (appSockAddr appData)))
           (source, sink) <- createClientHandler logId mb
           _ <- fork WorkerThread ("SendThread_" ++ show logId) $
                    source $= serializeServerMsg logId $$ appSink appData
           appSource appData $= parseClientMsg logId $$ sink

brokerServerMain :: IO ()
brokerServerMain =
    do args <- getArgs
       _ <- withLoggingAndLevel args INFO doWork
       return ()
    where
      doWork args =
          case args of
            (portStr : storeDir : rest@(_ : _))
                 | Just p <- readMay portStr ->
                             do mb <- createLocalBroker (LogId "server") storeDir
                                          (parseQueueSpec rest)
                                startBrokerServer p mb
            _ ->
                 do progName <- getProgName
                    logError
                        ("USAGE: " ++ progName ++ " PORT STORE_DIR QUEUE_SPEC...\n\n" ++
                         "QUEUE_SPEC is the name of a queue, possibly prefixed with " ++
                         "'@' to denote a persistent queue")
      parseQueueSpec l =
          flip map l $ \s ->
              case s of
                '@':queueName -> (QueueName (T.pack queueName), persistentQueueOpts)
                queueName -> (QueueName (T.pack queueName), transientQueueOpts)
