{-# OPTIONS_GHC -F -pgmF htfpp #-}
{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE CPP #-}
{-# LANGUAGE OverloadedStrings #-}
module Mgw.MessageQueue.BrokerStub
    (
      sendClientMain, recvClientMain
    , htf_thisModulesTests
    )
where

#include "src/macros.h"

----------------------------------------
-- LOCAL
----------------------------------------
import Mgw.MessageQueue.BrokerServer
import Mgw.MessageQueue.LocalBroker
import Mgw.MessageQueue.Protocol
import Mgw.MessageQueue.Types
import Mgw.Util.Concurrent
import Mgw.Util.Exception
import Mgw.Util.STM
import Mgw.Util.Tx
import Mgw.Util.Logging hiding (withLogId)
import Mgw.Util.Preview
import Mgw.Util.Sleep
import Mgw.Util.TestHelper
import Mgw.Util.TimeSpan

----------------------------------------
-- SITE-PACKAGES
----------------------------------------
import Data.Conduit.Network
import Data.HashSet (HashSet)
import Safe
import Test.Framework
import qualified Data.Conduit as C
import qualified Data.HashSet as HashSet
import qualified Data.Text as T
import qualified Data.Text.Encoding as T
import qualified Data.Vector as V

----------------------------------------
-- STDLIB
----------------------------------------
import Control.Monad
import Control.Monad.IO.Class
import System.Environment

connectChans ::
    LogId
    -> C.Source IO a
    -> C.Sink b IO ()
    -> TBMChan a
    -> TBMChan b
    -> IO (IO ())
connectChans logId source sink recvChan sendChan =
    do tid1 <- forkKillableThreadIO WorkerThread ("BrokerStubSender_" ++ show logId)
                    doSend
       tid2 <- forkKillableThreadIO WorkerThread ("BrokerStubReceiver_" ++ show logId)
                    doReceive
       return (mapM_ killThreadIO [tid1, tid2])
    where
      doSend = sendSource C.$$ sink
      doReceive = ignoreIOException (source C.$$ recvSink)
      sendSource =
          let loop = do mClMsg <- liftIO $ runTx $ readTBMChan sendChan
                        case mClMsg of
                          Just clMsg -> C.yield clMsg >> loop
                          Nothing -> return ()
          in loop
      recvSink =
          let loop = do mSrvMsg <- C.await
                        case mSrvMsg of
                          Just srvMsg ->
                              do liftIO $ runTx $ writeTBMChan recvChan srvMsg
                                 loop
                          Nothing -> liftIO $ runTx $ closeTBMChan recvChan
          in loop

createBrokerStub ::
    LogId
    -> (TBMChan ServerMessage -> TBMChan ClientMessage -> IO (IO ()))
    -> IO (MessageBroker LocalQueue)
createBrokerStub logId connectChans =
    do sendChan <- runTx $ newTBMChan ("BrokerStubSendChan_" ++ show logId) 20
       recvChan <- runTx $ newTBMChan ("BorkerStubRecvChan_" ++ show logId) 20
       kill <- connectChans recvChan sendChan
       logInfo (logMsg "waiting for queue list..")
       mMsg <- runTx $ readTBMChan recvChan
       case mMsg of
         Just (ServerQueues queues) ->
             do logInfo (logMsg ("received queue list"))
                run sendChan recvChan (V.toList queues)
         _ ->
             do kill
                logAndFail (logMsg ("Expected queue list from server but got something else: " ++
                                    show mMsg))
    where
      logMsg = withLogId logId
      localLogId = LogId (unLogId logId `T.append` "-local")
      run sendChan recvChan queues =
          do localBroker <- createLocalBroker localLogId "/tmp"
                                (map (\q -> (q, transientQueueOpts)) queues)
             serverSubscriptionsVar <- newTVarIO HashSet.empty
             closeVar <- newTVarIO False
             _ <- fork WorkerThread ("BrokerStubReceiveThread_" ++ show logId) $
                      handleMessage localBroker closeVar recvChan
             return $ MessageBroker
                  { mb_queues = mb_queues localBroker
                  , mb_subscribeToQueue = subscribe serverSubscriptionsVar closeVar
                                              sendChan localBroker
                  , mb_publishMessage = publish closeVar sendChan
                  }
      subscribe :: TVar (HashSet QueueName) -> TVar Bool -> TBMChan ClientMessage
                -> MessageBroker LocalQueue -> LocalQueue -> Subscriber -> IO ()
      subscribe serverSubscriptionsVar closeVar sendChan localBroker localQueue sub =
          do failIfClosed closeVar
             let queueName = lq_name localQueue
             mb_subscribeToQueue localBroker localQueue sub
             runTx $
                   do set <- readTVar serverSubscriptionsVar
                      unless (queueName `HashSet.member` set) $
                             do writeTVar serverSubscriptionsVar $! (HashSet.insert queueName set)
                                logInfo (logMsg ("Sending subscription for " ++ preview queueName
                                                 ++ " to server"))
                                writeTBMChan sendChan (ClientSubscribe queueName)
      publish closeVar sendChan localQueue msg =
          do failIfClosed closeVar
             logInfo (logMsg ("Sending message " ++ preview (msg_id msg) ++ " for queue " ++
                              preview (lq_name localQueue) ++ " to server"))
             runTx $ writeTBMChan sendChan (ClientPublishMessage (lq_name localQueue) msg)
      failIfClosed closeVar =
          do close <- runTx $ readTVar closeVar
             when close (logAndFail (logMsg ("Connection to server definitely closed")))
      handleMessage localBroker closeVar recvChan =
          let loop =
                  do mMsg <- runTx $ readTBMChan recvChan
                     case mMsg of
                       Nothing ->
                           do logInfo (logMsg "Received channel closed")
                              runTx $ writeTVar closeVar True
                       Just (ServerQueues _) ->
                           do logWarn (logMsg "Got ServerExistingQueues message, ignoring")
                              loop
                       Just (ServerPublishMessage queueName msg) ->
                           do case mb_lookupQueue localBroker queueName of
                                Nothing ->
                                    logWarn (logMsg ("Received message " ++ preview (msg_id msg) ++
                                                     " for unkown queue " ++ preview queueName))
                                Just queue ->
                                     do logInfo (logMsg ("Received message " ++
                                                         preview (msg_id msg) ++
                                                         " for queue " ++ preview queueName))
                                        mb_publishMessage localBroker queue msg
                              loop
          in loop

withBrokerClient :: String -> Int -> (MessageBroker LocalQueue -> IO ()) -> IO ()
withBrokerClient host port action =
    runTCPClient (clientSettings port hostBytes) $ \app ->
        do let logId = LogId (T.pack (show (appSockAddr app)))
               source = appSource app C.$= parseServerMsg logId
               sink = serializeClientMsg logId C.=$ appSink app
           mb <- createBrokerStub logId (connectChans logId source sink)
           action mb
    where
      hostBytes = T.encodeUtf8 (T.pack host)

sendClientMain :: IO ()
sendClientMain =
    do args <- getArgs
       _ <- withLoggingAndLevel args INFO doWork
       return ()
    where
      doWork args =
          case args of
            [host, portStr, queueName, msgId, msg]
                | Just port <- readMay portStr ->
                    withBrokerClient host port $ \mb ->
                       case mb_lookupQueue mb (QueueName (T.pack queueName)) of
                         Nothing -> logError ("Unkown queue: " ++ queueName)
                         Just q ->
                             do let bytes = T.encodeUtf8 (T.pack msg)
                                mb_publishMessage mb q (Message (MessageId (T.pack msgId)) bytes)
                                sleepTimeSpan (milliseconds 50)
            _ ->
                do progName <- getProgName
                   logError ("USAGE: " ++ progName ++ " HOST PORT QUEUE MSG_ID MSG")


recvClientMain :: IO ()
recvClientMain =
    do args <- getArgs
       _ <- withLoggingAndLevel args INFO doWork
       return ()
    where
      doWork args =
          case args of
            [host, portStr, queueName]
                | Just port <- readMay portStr ->
                    withBrokerClient host port $ \mb ->
                       case mb_lookupQueue mb (QueueName (T.pack queueName)) of
                         Nothing -> logError ("Unkown queue: " ++ queueName)
                         Just q ->
                             do mb_subscribeToQueue mb q (mkSubscriber "recvClient" handleMsg)
                                wait
            _ ->
                do progName <- getProgName
                   logError ("USAGE: " ++ progName ++ " HOST PORT QUEUE")
      handleMsg msg =
          let bytes = msg_payload msg
          in case T.decodeUtf8' bytes of
               Right t -> logNote ("Received message: " ++ T.unpack t)
               Left _ -> logNote ("Received binary message")
      wait =
          do logNote ("Hit ENTER to terminate the program")
             _ <- getLine
             return ()

brokerStubWithBrokerServerTest ::
    (MessageBroker LocalQueue -> MessageBroker LocalQueue -> IO ()) -> IO ()
brokerStubWithBrokerServerTest fun =
    do serverBroker <- createLocalBroker (LogId "server-local") "/tmp" queues
       (source, sink) <- createClientHandler (LogId "server") serverBroker
       localBroker <- createBrokerStub (LogId "stub")
                          (connectChans (LogId "stub-chans") source sink)
       fun serverBroker localBroker
    where
      queues = [(testQueue1, transientQueueOpts)
               ,(testQueue2, transientQueueOpts)]

test_brokerStubWithBrokerServer1 :: IO ()
test_brokerStubWithBrokerServer1 =
    brokerStubWithBrokerServerTest $ \_serverBroker localBroker ->
       messageBrokerTest localBroker localBroker

test_brokerStubWithBrokerServer2 :: IO ()
test_brokerStubWithBrokerServer2 =
    brokerStubWithBrokerServerTest $ \serverBroker localBroker ->
       messageBrokerTest serverBroker localBroker

test_brokerStubWithBrokerServer3 :: IO ()
test_brokerStubWithBrokerServer3 =
    brokerStubWithBrokerServerTest $ \serverBroker localBroker ->
       messageBrokerTest localBroker serverBroker

test_brokerStubWithBrokerServer4 :: IO ()
test_brokerStubWithBrokerServer4 =
    do serverBroker <- createLocalBroker (LogId "server-local") "/tmp" queues
       (source1, sink1) <- createClientHandler (LogId "server1") serverBroker
       localBroker1 <- createBrokerStub (LogId "stub1")
                           (connectChans (LogId "stub1-chans") source1 sink1)
       (source2, sink2) <- createClientHandler (LogId "server2") serverBroker
       localBroker2 <- createBrokerStub (LogId "stub2")
                           (connectChans (LogId "stub2-chans") source2 sink2)
       messageBrokerTest localBroker1 localBroker2
    where
      queues = [(testQueue1, transientQueueOpts)
               ,(testQueue2, transientQueueOpts)]
